(*====================================================================

<***MODULE>confTheory<***/MODULE>

<***DESC NAME="ROOT"> 
This theory defines the notion of predicate confinement and provides
some theorems about properties of confinement.
<***/DESC>

<***AUTHOR> 
@@NAME Wishnu Prasetya
@@EMAIL wishnu@cs.uu.nl
@@EMAIL conjurerPrime@yahoo.com
@@URL www.cs.uu.nl/~wishnu
<***/AUTHOR>

<***VERSION>
@@DATE Feb 2002 on Hol2.00 Taupo 6
@@DATE Sept 2001
<***/VERSION>

====================================================================*)

open HolKernel Parse basicHol90Lib ;
infix THEN THENL THENC ORELSE ORELSEC THEN_TCL ORELSE_TCL ## |-> ;
infixr --> ;


val _ = new_theory "conf" ;

(* 
load "predTheory" 		;
load "predLib"			;
*)

open predTheory predLib 	;

(*
val State  = ty_antiq (Type `:'var -> 'val`) ;
val Pred   = ty_antiq (Type `:^(type_of State) -> bool`) ;
*)

(*====================================================================

<***DESC name="Partial equality on functions">

Given two functions f and g from a to b and a predicate V which is a
subset of a, the two functions are partialy equal over V if they are
equal on values in V. 

<blockquote>
<pre>p_EQ_def : |- !V f g.  p_EQ V f  g  =   (!x. V x ==> (f x = g x))</pre>
</blockquote>

<***/DESC>

====================================================================*)

val p_EQ_def
	=
	new_definition ("p_EQ_def",
	--`p_EQ V f g = (!x. V x ==> (f x = g x))`--		)	;


(*
val _ = add_rule 
	{ term_name 	= "p_EQ",
	  fixity 	= TruePrefix 524,
	  pp_elements	= [PPBlock ([TOK "ON", BreakSpace(1,2), TM, 
				     BreakSpace(1,0), 
				     TOK ":::"], 
				     (PP.CONSISTENT, 3)),
			   BreakSpace(1,2), TM,  BreakSpace(1,0),
			   BeginFinalBlock(PP.CONSISTENT,2),
			   TOK "===",  BreakSpace(1,0)],
	  paren_style	= OnlyIfNecessary,
	  block_style	= (AroundEachPhrase, (PP.INCONSISTENT,0))
	} ;
*)


(*====================================================================

<***DESC name="conf">

For a state predicate p and a set of variables V we write <tt>p
conf V</tt> to say that p only restricts the values of variables in V,
or in other words: p is confined by V.

<PRE>conf_def : |- !p V. p conf V = (!s t. p_EQ V s t ==> (p s = p t))</PRE>

Parsed as infixl 522.

<***/DESC>

====================================================================*)

val conf_def 
	= 
	new_infixl_definition ("conf_def",
	--`$conf p V = (!s t. p_EQ V s t ==> (p s = p t))`--,
	522
	) ;

(*====================================================================

<***DESC name="Confinement of TT and FF">

TT and FF are confined by empty set (FF), which implies that they are
confined by any set (next theorem).

<pre>TT_conf_thm	:	TT conf FF
FF_conf_thm	:	FF conf FF</pre>

<***/DESC>

====================================================================*)

val TT_conf_thm
	=
	store_thm ("TT_conf_thm",
	--`TT conf FF`--,
	REWRITE_TAC [p_EQ_def, conf_def]
	THEN valid_PROVE_TAC) ;

val FF_conf_thm
	=
	store_thm ("FF_conf_thm",
	--`TT conf FF`--,
	REWRITE_TAC [p_EQ_def, conf_def]
	THEN valid_PROVE_TAC) ;

(*====================================================================

<***DESC name="Monotonicity of confinement">

The following theorem states if a predicate if confined by V, then it
is confined by any super set of V. The superset relation here is
represented by IMP.

<pre>conf_MONO_thm : valid (V IMP W) /\ p conf V ==> p conf W </pre>

<***/DESC>

====================================================================*)

val conf_MONO_thm
	=
	store_thm ("conf_MONO_thm",
	--`valid (V IMP W) /\ p conf V ==> p conf W`--,
	REWRITE_TAC [p_EQ_def, conf_def]
	THEN valid_PROVE_TAC) ;

(*====================================================================

<***DESC name="Closedness of confiment">

The following set of theorems state that confinement is closed under
all predicate operators.

<pre>conf_AND_CLOSED_thm 	:	p conf V /\ q conf V ==> p AND   q conf V 
conf_OR_CLOSED_thm 	:	p conf V /\ q conf V ==> p OR    q conf V 
conf_IMP_CLOSED_thm 	:	p conf V /\ q conf V ==> p IMP   q conf V 
conf_EQUAL_CLOSED_thm 	:	p conf V /\ q conf V ==> p EQUAL q conf V 
conf_NOT_CLOSED_thm 	:	p conf V             ==> NOT p     conf V 
conf_qAND_CLOSED_thm 	:	(!i. W i ==> P i conf V) ==> (!!i::W. P i) conf V
conf_qOR_CLOSED_thm	:	(!i. W i ==> P i conf V) ==> (??i::W. P i) conf V</pre>

<***/DESC>

====================================================================*)


val conf_AND_CLOSED_thm
	=
	store_thm ("conf_AND_CLOSED_thm",
	--`p conf V /\ q conf V ==> p AND q conf V`--,
	REWRITE_TAC [p_EQ_def, conf_def]
	THEN valid_PROVE_TAC) ;

val conf_OR_CLOSED_thm
	=
	store_thm ("conf_OR_CLOSED_thm",
	--`p conf V /\ q conf V ==> p OR q conf V`--,
	REWRITE_TAC [p_EQ_def, conf_def]
	THEN valid_PROVE_TAC) ;

val conf_IMP_CLOSED_thm
	=
	store_thm ("conf_IMP_CLOSED_thm",
	--`p conf V /\ q conf V ==> p IMP q conf V`--,
	REWRITE_TAC [p_EQ_def, conf_def]
	THEN valid_PROVE_TAC) ;

val conf_EQUAL_CLOSED_thm
	=
	store_thm ("conf_EQUAL_CLOSED_thm",
	--`p conf V /\ q conf V ==> p EQUAL q conf V`--,
	REWRITE_TAC [p_EQ_def, conf_def]
	THEN valid_PROVE_TAC) ;

val conf_NOT_CLOSED_thm
	=
	store_thm ("conf_NOT_CLOSED_thm",
	--`p conf V ==> NOT p conf V`--,
	REWRITE_TAC [p_EQ_def, conf_def]
	THEN valid_PROVE_TAC) ;

val conf_qAND_CLOSED_thm
	=
	store_thm ("conf_qAND_CLOSED_thm",
	--`(!i. W i ==> P i conf V) ==> (!!i::W. P i) conf V`--,
	REWRITE_TAC [p_EQ_def, conf_def]
	THEN valid_PROVE_TAC) ;

val conf_qOR_CLOSED_thm
	=
	store_thm ("conf_qOR_CLOSED_thm",
	--`(!i. W i ==> P i conf V) ==> (??i::W. P i) conf V`--,
	REWRITE_TAC [p_EQ_def, conf_def]
	THEN valid_PROVE_TAC) ;


val _ = export_theory() ;



