(*====================================================================

<***MODULE>nonInterferenceTheory<***/MODULE>

<***DESC NAME="ROOT"> 
...
<***/DESC>

<***AUTHOR> 
@@NAME Wishnu Prasetya
@@EMAIL wishnu@cs.uu.nl
@@EMAIL conjurerPrime@yahoo.com
@@URL www.cs.uu.nl/~wishnu
<***/AUTHOR>

<***VERSION>
@@DATE Feb 2002 on Hol2.00 Taupo 6
@@DATE Sept 2001
<***/VERSION>


====================================================================*)

open HolKernel Parse basicHol90Lib ;
infix THEN THENL THENC ORELSE ORELSEC THEN_TCL ORELSE_TCL ## |-> ;
infixr --> ;


(*
load "predTheory" ;
load "actionTheory" ;
load "confTheory" ;
load "uprogTheory" ;
load "bossLib" ;
load "predLib" ;
*)

open predTheory actionTheory confTheory uprogTheory bossLib predLib ;

val _ = new_theory "nonInterference" ;

(*
val State  = ty_antiq (Type `:'var -> 'val`) ;
val Pred   = ty_antiq (Type `:^(type_of State) -> bool`) ;
*)

val UNDISCH_ALL_TAC =
	let
	fun ttac thm = UNDISCH_TAC (concl thm)
	in
	EVERY_ASSUM ttac
	end ;

(*====================================================================

<***DESC NAME="Pset_def">

The Pset of an action e is the set of all predicates which are
preserved by the action. <p>

<pre>@@CODEREF "Pset_def"</pre>

<***/DESC>
====================================================================*)

val Pset_def
	=

(* <***CODE name="Pset_def"> *)

	new_definition("Pset_def",
	--`Pset e p = HOA p e p`-- ) ;

(* <***/CODE> *)

val conf_Pset_thm
	=
	store_thm ("conf_Pset_thm",
	--`p conf V = Pset (p_EQ V) p`--,

	RW_TAC std_ss [conf_def, Pset_def, HOA_def, p_EQ_def]
	THEN PROVE_TAC []) ;

val confset_def
	=
	new_definition("confset_def",
	--`confset V = (\p. p conf V)`-- ) ;

val p_EQ_adef 
	= 
	store_thm ("p_EQ_adef",
	--`p_EQ V = (\f g. !x. V x ==> (f x = g x))`--,
	REPEAT ((CONV_TAC o CHANGED_CONV o DEPTH_CONV) FUN_EQ_CONV)
	THEN RW_TAC std_ss [p_EQ_def]) ;	

val SUSPEND_def
	=
	new_definition ("SUSPEND_def",
	--`SUSPEND Q J A p q
	   =
	   STABLE Q J  /\
	   (!p' a. Q a /\ A p'
		==> 
		HOA (J AND p AND p' AND NOT q) 
		    a 
		    ((p AND p') OR q))`--
	) ;

val IMPLEMENT_def 
	=
	new_definition ("IMPLEMENT_def",
	--`IMPLEMENT Q J e p q
	   =
	   SUSPEND Q J (Pset e) p q`--) ;

val PRES_def 
	=
	new_definition ("PRES_def",
	--`PRES Q J V p q
	   =
	   SUSPEND Q J (confset V) p q`--) ;

val PRES_adef 
	=
	store_thm ("PRES_adef",
	--`PRES Q J V p q = IMPLEMENT Q J (p_EQ V) p q`--,
	RW_TAC std_ss [confset_def, conf_Pset_thm, PRES_def, IMPLEMENT_def]) ;
	
val CHOICE_def 
	= 
	new_infixl_definition ("CHOICE_def",
	--`$CHOICE a b = (\s. a s OR b s)`--,
	500) ;

val SYNC_def 
	= 
	new_infixl_definition ("SYNC_def",
	--`$SYNC a b = (\s. a s AND b s)`--,
	504) ;

val level0_defs = [Pset_def, conf_def, confset_def, p_EQ_adef] ;
val level1_defs = [HOA_def, SUSPEND_def, STABLE_def] ;

val confset_MONO_thm
	=
	store_thm ("confset_MONO_thm",
	--`valid (confset V IMP confset (V OR W))`--,
	REWRITE_TAC ([valid_def, pIMP_def] @ level0_defs @ level1_defs)	
	THEN valid_PROVE_TAC) ;
	
val Pset_MONO_thm
	=
	store_thm ("Pset_MONO_thm",
	--`valid (Pset e IMP Pset (e SYNC f))`--,
	REWRITE_TAC ([valid_def,SYNC_def,pIMP_def] @ level0_defs @ level1_defs)	
	THEN valid_PROVE_TAC) ;
	
val Suspension_Shrinking_thm
	=
	store_thm ("Suspension_Shrinking_thm",
	--`SUSPEND Q J B p q 	/\
	   valid (A IMP B)
	   ==>
	   SUSPEND Q J A p q`--,
	REWRITE_TAC level1_defs
	THEN valid_PROVE_TAC ) ;

val Implementation_Shrinking_thm
	=
	store_thm ("Implementation_Shrinking_thm",
	--`IMPLEMENT Q J (e SYNC f) p q
	   ==>
	   IMPLEMENT Q J e p q`--,
	PROVE_TAC [IMPLEMENT_def, Suspension_Shrinking_thm, Pset_MONO_thm]
	) ;

val Preservation_Shrinking_thm
	=
	store_thm ("Preservation_Shrinking_thm",
	--`PRES Q J (V OR W) p q
	   ==>
	   PRES Q J V p q`--,
	PROVE_TAC [PRES_def, Suspension_Shrinking_thm, confset_MONO_thm]
	) ;

(* first version: too restrictive!!

val isPartial_CrossConj_def
	=
	new_definition ("Cross_Conjunction_def",
	--`isPartial_CrossConj C A B
	   =
	   !p. C p ==> ?p1 p2. A p1 /\ B p2 /\ (p = p1 AND p2)`--) ;

*)

val isCoercion_def
	=
	new_definition ("isCoercion_def",
	--`isCoercion C coerceToC = !s. C (coerceToC s)`-- ) ;

val isClosureCoercion_def 
	=
	new_definition ("isClosureCoercion_def",
	--`isClosureCoercion coerce
	   =
	   !s. coerce s s`--) ;

val isCoercionClosed_def 
	=
	new_definition ("isCoercionClosed_def",
	--`isCoercionClosed C coerceToC  
	   =
	   !p s. C p /\ p s ==> valid (coerceToC s IMP p)`--) ;

val coercion_closedness_thm
	=
	store_thm ("coercion_closedness_thm",
	--`isClosureCoercion coerceToC /\ 
	   isCoercionClosed C coerceToC /\
	   C p
	   ==>
	   (p = ??s::p. coerceToC s)`--,
	REWRITE_TAC [isClosureCoercion_def, isCoercionClosed_def]
	THEN valid_PROVE_TAC) ;

val isCoercion_CrossConj_def 
	=
	new_definition ("isCoercion_CrossConj_def",
	--`isCoercion_CrossConj toC toA toB C
	   =
	   !p s. C p /\ p s ==> (toC s = toA s AND toB s)`--
	) ;

val isPartial_CrossConj_def
	=
	new_definition ("Cross_Conjunction_def",
	--`isPartial_CrossConj C A B
	   =
	   ?coerceToC coerceToA coerceToB.
	      isClosureCoercion coerceToC /\	
	      isCoercion A coerceToA /\
	      isCoercion B coerceToB /\
	      isCoercionClosed C coerceToC /\
	      isCoercion_CrossConj coerceToC coerceToA coerceToB C
	`--
	) ;

val lemma1
	=
	prove (
	--`(!s. p0 s ==>
		HOA (J AND p1 AND f s AND NOT q1) 
		    a 
	            (p1 AND f s OR q1)
		/\
		HOA (J AND p2 AND g s AND NOT q2) 
		    a 
	            (p2 AND g s OR q2))
	  ==>
	  HOA (J AND (p1 AND p2) AND (??s::p0. f s AND g s) AND NOT (q1 OR q2))
	      a
	      ((p1 AND p2) AND (??s::p0. f s AND g s) OR (q1 OR q2))
	`--,
	REWRITE_TAC [HOA_def]
	THEN valid_PROVE_TAC) ;	

val Suspension_Conjunction_1_thm
	=
	store_thm ("Suspension_Conjunction_1_thm",
	--`SUSPEND Q J A p1 (q1:'a->bool)	/\
	   SUSPEND Q J B p2 q2	/\
	   isPartial_CrossConj C A B
	   ==>
	   SUSPEND Q J C (p1 AND p2) (q1 OR q2)`--,

	RW_TAC std_ss [SUSPEND_def, isPartial_CrossConj_def, 
		       isCoercion_CrossConj_def, isCoercion_def]

	THEN SUBGOAL_THEN
	      (--`(p':'a->bool) =  ??s::p'. coerceToC s`--)
	      SUBST1_TAC
	THENL [ (* this part cannot be solved directly with PROVE_TAC *)
	        MATCH_MP_TAC (GEN_ALL coercion_closedness_thm)
		THEN PROVE_TAC [],
		ALL_TAC ]

	THEN SUBGOAL_THEN
		(--`((??s::(p':'a->bool). coerceToC s) : 'a->bool)
		    =
	            (??s::p'. coerceToA s AND coerceToB s)`--)
	       SUBST1_TAC
	THENL [	RW_TAC std_ss [RES_qOR_def]
	        THEN CONV_TAC (DEPTH_CONV FUN_EQ_CONV)
		THEN BETA_TAC
	        THEN PROVE_TAC [],
		ALL_TAC ]

	THEN RW_TAC std_ss [lemma1]
	) ;


val Suspension_Conjunction_2_thm
	=
	store_thm ("Suspension_Conjunction_2_thm",
	--`SUSPEND Q J A p q	/\
	   SUSPEND Q J B p q	/\
	   isPartial_CrossConj C A B
	   ==>
	   SUSPEND Q J C p q`--,
	let
	val lem1 = prove (--`p AND p = p`--, pred_eq_PROVE_TAC) ;
	val lem2 = prove (--`p OR  p = p`--, pred_eq_PROVE_TAC) ;
	in
	PROVE_TAC [Suspension_Conjunction_1_thm, lem1, lem2]
	end
	) ;

val isClosureCoercion_p_EQ_thm
	=
	store_thm ("isClosureCoercion_p_EQ_thm",
	--`isClosureCoercion (p_EQ V)`--,
	REWRITE_TAC [isClosureCoercion_def, p_EQ_adef]
	THEN valid_PROVE_TAC) ;

val isCoercion_p_EQ_thm
	=
	store_thm ("isCoercion_p_EQ_thm",
	--`isCoercion (confset V) (p_EQ V)`--,
	REWRITE_TAC [isCoercion_def, confset_def, conf_def, p_EQ_adef]
	THEN valid_PROVE_TAC) ;

val isCoercionClosed_p_EQ_thm
	=
	store_thm("isCoercionClosed_p_EQ_thm",
	--`isCoercionClosed (confset V) (p_EQ V)`--,
	REWRITE_TAC [isCoercionClosed_def, confset_def, conf_def, p_EQ_adef]
	THEN valid_PROVE_TAC) ;

val isCoercion_CrossConj_p_EQ_thm
	=
	store_thm ("isCoercion_CrossConj_p_EQ_thm",
	--`isCoercion_CrossConj 
		(p_EQ (V OR W)) 
		(p_EQ V) 
		(p_EQ W)
		(confset (V OR W))`--,
	REWRITE_TAC [isCoercion_CrossConj_def, confset_def, conf_def, p_EQ_adef]
	THEN valid_PROVE_TAC) ;
	
val isPartial_CrossConj_confset_thm
	=
	store_thm ("isPartial_CrossConj_confset_thm",
	--`isPartial_CrossConj 
		(confset (V OR W)) 
		(confset (V:'var->bool)) 
		(confset W)`--,
	PROVE_TAC [isPartial_CrossConj_def,isClosureCoercion_p_EQ_thm,
		   isCoercion_p_EQ_thm,
		   isCoercionClosed_p_EQ_thm,
		   isCoercion_CrossConj_p_EQ_thm ] 
	) ;

	
val coerceToPset_def
	=
	new_definition ("coerceToPset_def",
	--`coerceToPset e s
	   =
	   (\t. (!p. Pset e p /\ p s ==> p t))`-- ) ;

val isClosureCoercion_coerceToPset_thm
	=
	store_thm
	("isClosureCoercion_coerceToPset_thm",
	--`isClosureCoercion (coerceToPset e)`--,
	REWRITE_TAC [isClosureCoercion_def, coerceToPset_def, Pset_def, HOA_def]
	THEN valid_PROVE_TAC) ;

val isCoercion_coerceToPset_thm
	=
	store_thm ("isCoercion_coerceToPset_thm",
	--`isCoercion (Pset e) (coerceToPset e)`--,
	REWRITE_TAC [isCoercion_def, coerceToPset_def, Pset_def, HOA_def]
	THEN valid_PROVE_TAC) ;

val isCoercionClosed_coerceToPset_thm
	=
	store_thm("isCoercionClosed_coerceToPset_thm",
	--`isCoercionClosed (Pset e) (coerceToPset e)`--,
	REWRITE_TAC [isCoercionClosed_def, coerceToPset_def, Pset_def, HOA_def]
	THEN valid_PROVE_TAC) ;

(* 	

However, coerceToPset does not in general satisfy
isCoercion_CrossConj, and therefore e SYNC f will not satisfy the
isPartial_CrossConj relation either.

It is however reasonable to assume that there are reasonable
conditions to make coerceToPset satisfies isCoercion_CrossConj.

So, while the suspension conjunction principle will hold for
preservation too, it won't hold for e SYNC f unless we put some extra
restrictions on e and f.

*)


val Preservation_Conjunction_1_thm
	=
	store_thm ("Preservation_Conjunction_1_thm",
	--`PRES Q J V p1 q1	/\
	   PRES Q J W p2 q2	
	   ==>
	   PRES Q J (V OR W) (p1 AND p2) (q1 OR q2)`--,

	PROVE_TAC [PRES_def, isPartial_CrossConj_confset_thm,
		   Suspension_Conjunction_1_thm]
	) ;

val Preservation_Conjunction_2_thm
	=
	store_thm ("Preservation_Conjunction_2_thm",
	--`PRES Q J V p q	/\
	   PRES Q J W p q	
	   ==>
	   PRES Q J (V OR W) p q`--,
	PROVE_TAC [PRES_def, isPartial_CrossConj_confset_thm,
		   Suspension_Conjunction_2_thm]
	) ;


val Suspension_JStrenthening_thm
	=
	store_thm ("Suspension_JStrenthening_thm",
	--`valid (J2 IMP J1) 			/\
	   STABLE Q J2			/\ 
	   SUSPEND Q J1 A p q	
	   ==>
	   SUSPEND Q J2 A p q`--,
	REWRITE_TAC level1_defs
	THEN valid_PROVE_TAC) ;	

val Implementation_JStrenthening_thm
	=
	store_thm ("Implementation_JStrenthening_thm",
	--`valid (J2 IMP J1) 			/\
	   STABLE Q J2			/\ 
	   IMPLEMENT Q J1 e p q	
	   ==>
	   IMPLEMENT Q J2 e p q`--,
	PROVE_TAC [Suspension_JStrenthening_thm, IMPLEMENT_def]) ;

val Preservation_JStrenthening_thm
	=
	store_thm ("Preservation_JStrenthening_thm",
	--`valid (J2 IMP J1) 			/\
	   STABLE Q J2			/\ 
	   PRES Q J1 V p q	
	   ==>
	   PRES Q J2 V p q`--,
	PROVE_TAC [Suspension_JStrenthening_thm, PRES_def]) ;

val Suspension_JWeakening_thm
	=
	store_thm ("Suspension_JWeakening_thm",
	--`SUSPEND Q J1 A p q	/\
	   SUSPEND Q J2 A p q	
	   ==>
	   SUSPEND Q (J1 OR J2) A p q`--,
	REWRITE_TAC level1_defs
	THEN valid_PROVE_TAC) ;	

val Implementation_JWeakening_thm
	=
	store_thm ("Implementation_JWeakening_thm",
	--`IMPLEMENT Q J1 e p q	/\
	   IMPLEMENT Q J2 e p q	
	   ==>
	   IMPLEMENT Q (J1 OR J2) e p q`--,
	PROVE_TAC [Suspension_JWeakening_thm, IMPLEMENT_def]) ;

val Preservation_JWeakening_thm
	=
	store_thm ("Preservation_JWeakening_thm",
	--`PRES Q J1 V p q	/\
	   PRES Q J2 V p q	
	   ==>
	   PRES Q (J1 OR J2) V p q`--,
	PROVE_TAC [Suspension_JWeakening_thm, PRES_def]) ;

val Suspension_Substitution_thm
	=
	store_thm ("Suspension_Substitution_thm",
	--`valid (J IMP (p EQUAL p'))			/\
	   valid (J IMP (q EQUAL q'))			/\
	   SUSPEND Q J A p q	
	   ==>
	   SUSPEND Q J A p' q'`--,
	REWRITE_TAC level1_defs
	THEN valid_PROVE_TAC) ;	

val Implemetation_Substitution_thm
	=
	store_thm ("Implemetation_Substitution_thm",
	--`valid (J IMP (p EQUAL p'))			/\
	   valid (J IMP (q EQUAL q'))			/\
	   IMPLEMENT Q J e p q	
	   ==>
	   IMPLEMENT Q J e p' q'`--,
	PROVE_TAC [Suspension_Substitution_thm, IMPLEMENT_def]) ;

val Preservation_Substitution_thm
	=
	store_thm ("Preservation_Substitution_thm",
	--`valid (J IMP (p EQUAL p'))			/\
	   valid (J IMP (q EQUAL q'))			/\
	   PRES Q J V p q	
	   ==>
	   PRES Q J V p' q'`--,
	PROVE_TAC [Suspension_Substitution_thm, PRES_def]) ;

val Suspension_PostWeakening_thm
	=
	store_thm ("Suspension_PostWeakening_thm",
	--`valid (J AND q IMP  q')			/\
	   SUSPEND Q J A p q	
	   ==>
	   SUSPEND Q J A p q'`--,
	REWRITE_TAC level1_defs
	THEN valid_PROVE_TAC) ;	

val Implementation_PostWeakening_thm
	=
	store_thm ("Implementation_PostWeakening_thm",
	--`valid (J AND q IMP  q')			/\
	   IMPLEMENT Q J e p q	
	   ==>
	   IMPLEMENT Q J e p q'`--,
	PROVE_TAC [Suspension_PostWeakening_thm, IMPLEMENT_def]) ;

val Preservation_PostWeakening_thm
	=
	store_thm ("Preservation_PostWeakening_thm",
	--`valid (J AND q IMP  q')			/\
	   PRES Q J V p q	
	   ==>
	   PRES Q J V p q'`--,
	PROVE_TAC [Suspension_PostWeakening_thm, PRES_def]) ;

val Suspension_UnlessConjunction_thm
	=
	store_thm ("Suspension_UnlessConjunction_thm",
	--`SUSPEND Q J A p q		/\
	   SUSPEND Q J A r s	
	   ==>
	   SUSPEND Q J A (p AND r) (q OR s)`--,
	REWRITE_TAC level1_defs
	THEN valid_PROVE_TAC) ;	

val Implementation_UnlessConjunction_thm
	=
	store_thm ("Implementation_UnlessConjunction_thm",
	--`IMPLEMENT Q J e p q		/\
	   IMPLEMENT Q J e r s	
	   ==>
	   IMPLEMENT Q J e (p AND r) (q OR s)`--,
	PROVE_TAC [Suspension_UnlessConjunction_thm, IMPLEMENT_def]) ;

val Preservation_UnlessConjunction_thm
	=
	store_thm ("Preservation_UnlessConjunction_thm",
	--`PRES Q J V p q		/\
	   PRES Q J V r s	
	   ==>
	   PRES Q J V (p AND r) (q OR s)`--,
	PROVE_TAC [Suspension_UnlessConjunction_thm, PRES_def]) ;

val Suspension_UnlessDisjunction_thm
	=
	store_thm ("Suspension_UnlessDisjunction_thm",
	--`SUSPEND Q J A p q		/\
	   SUSPEND Q J A r s	
	   ==>
	   SUSPEND Q J A (p OR r) (q OR s)`--,
	REWRITE_TAC level1_defs
	THEN valid_PROVE_TAC) ;	

val Implementation_UnlessDisjunction_thm
	=
	store_thm ("Implementation_UnlessDisjunction_thm",
	--`IMPLEMENT Q J e p q		/\
	   IMPLEMENT Q J e r s	
	   ==>
	   IMPLEMENT Q J e (p OR r) (q OR s)`--,
	PROVE_TAC [Suspension_UnlessDisjunction_thm, IMPLEMENT_def]) ;

val Preservation_UnlessDisjunction_thm
	=
	store_thm ("Preservation_UnlessDisjunction_thm",
	--`PRES Q J V p q		/\
	   PRES Q J V r s	
	   ==>
	   PRES Q J V (p OR r) (q OR s)`--,
	PROVE_TAC [Suspension_UnlessDisjunction_thm, PRES_def]) ;


val Suspension_whenStrengthening_thm
	=
	store_thm ("Suspension_whenStrengthening_thm",
	--`A p' /\ SUSPEND Q J A p q		
	   ==>
	   SUSPEND Q J A (p AND p') q`--,
	REWRITE_TAC level1_defs
	THEN valid_PROVE_TAC) ;	

val Implementation_whenStrengthening_thm
	=
	store_thm ("Implementation_whenStrengthening_thm",
	--`Pset e p' /\ IMPLEMENT Q J e p q		
	   ==>
	   IMPLEMENT Q J e (p AND p') q`--,
	PROVE_TAC [Suspension_whenStrengthening_thm, IMPLEMENT_def]) ;

val Preservation_whenStrengthening_thm
	=
	store_thm ("Preservation_whenStrengthening_thm",
	--`p' conf V /\ PRES Q J V p q		
	   ==>
	   PRES Q J V (p AND p') q`--,
	PROVE_TAC [Suspension_whenStrengthening_thm, PRES_def, confset_def]) ;

val alwaysSUSPEND_def 
	=
	new_definition ("alwaysSUSPEND_def",
	--`alwaysSUSPEND Q A  =  SUSPEND Q TT A TT FF`--) ;

val noWriteAccess_def 
	=
	new_definition ("noWriteAccess_def",
	--`noWriteAccess Q V  =  PRES Q TT V TT FF`--) ;

val noWriteAccess_adef
	=
	store_thm ("noWriteAccess_adef",
	--`noWriteAccess Q V  = alwaysSUSPEND Q (confset V)`--,
	PROVE_TAC [noWriteAccess_def, alwaysSUSPEND_def, PRES_def]) ;


val Suspension_Extension_With_LocPred_thm 
	=
	store_thm ("Suspension_Extension_With_LocPred_thm",
	--`alwaysSUSPEND Q B 			/\ 
	   isPartial_CrossConj C A B		/\
	   valid (A IMP C)
	   ==>
	   (SUSPEND Q J A p (q:'a->bool)	
	    =
	    SUSPEND Q J C p q)`--,
	let
	val lem1 = prove (--`valid (J IMP TT)`--, valid_PROVE_TAC) 
	val lem2 = prove (--`p AND TT = p`--, pred_eq_PROVE_TAC)
	val lem3 = prove (--`p OR FF = p`--, pred_eq_PROVE_TAC)
	in
	RW_TAC std_ss [alwaysSUSPEND_def, Suspension_Shrinking_thm]
	THEN EQ_TAC
	THEN TRY (PROVE_TAC [Suspension_Shrinking_thm])
	THEN STRIP_TAC
	THEN SUBGOAL_THEN (--`STABLE Q (J:'a->bool)`--)
	     ASSUME_TAC
	THENL [ UNDISCH_ALL_TAC THEN RW_TAC std_ss [SUSPEND_def],
	        ALL_TAC ]
	THEN PROVE_TAC [Suspension_Conjunction_1_thm,
	                Suspension_JStrenthening_thm, lem1, lem2, lem3]
	end
	) ;

val confset_mono = prove (--`valid (confset V IMP confset (V OR W))`--,
	REWRITE_TAC [confset_def, conf_def, p_EQ_def]
	THEN valid_PROVE_TAC) ;

val Preservation_Extension_With_LocVars_thm 
	=
	store_thm ("Preservation_Extension_With_LocVars_thm",
	--`noWriteAccess Q W 			
	   ==>
	   (PRES Q J V p q
	    =
	    PRES Q J (V OR W) p q)`--,

	PROVE_TAC [PRES_def, Suspension_Extension_With_LocPred_thm,
		   isPartial_CrossConj_confset_thm, confset_mono,
		   noWriteAccess_adef ]
	) ;

val Suspension_whenStrengthening_With_LocPred_thm 
	=
	store_thm ("Suspension_whenStrengthening_With_LocPred_thm",
	--`isPartial_CrossConj C A B		/\
	   valid (A IMP C)			/\
	   valid (B IMP C)			/\
	   alwaysSUSPEND Q B			/\
	   B p'					/\
	   SUSPEND Q J A p q
	   ==>
	   SUSPEND Q J A (p AND p') q`--,
	let
	val lemma = prove(--`valid (B IMP C) = !q. B q ==> C q`--,
			valid_PROVE_TAC) 
	in
	PROVE_TAC [Suspension_Extension_With_LocPred_thm,
		   Suspension_whenStrengthening_thm, lemma]
	end
	) ;

val Preservation_whenStrengthening_With_LocVars_thm 
	=
	store_thm ("Preservation_whenStrengthening_With_LocVars_thm",
	--`noWriteAccess Q W			/\
	   p' conf W				/\
	   PRES Q J V p q
	   ==>
	   PRES Q J V (p AND p') q`--,
	let
	val lem1 = prove(--`a conf W = confset W a`--, RW_TAC std_ss [confset_def])
	val lem2 = prove(--`V OR W = W OR V`--, pred_eq_PROVE_TAC)
	in
	PROVE_TAC [PRES_def, Suspension_whenStrengthening_With_LocPred_thm,
		   isPartial_CrossConj_confset_thm, confset_mono,
		   lem1, lem2 , noWriteAccess_adef ]
	end
	) ;

val _ = export_theory() ;
