(* ---------------------------------------------------------------

<***MODULE>predLib<***/MODULE>

<***DESC NAME="ROOT">
This module provides some conversions to reduce embedded predicates to
ordinary HOL formulas. The module also provides some tactics to prove
tautology and equalities between embedded predicates.
<***/DESC>

<***AUTHOR> 
@@NAME Wishnu Prasetya
@@EMAIL wishnu@cs.uu.nl 
@@EMAIL conjurerPrime@yahoo.com
@@URL www.cs.uu.nl/~wishnu
<***/AUTHOR>

<***VERSION>
@@DATE Feb 2002 on Hol2.00 Taupo 6
@@DATE Sept 2001
<***/VERSION>

<***DESC NAME="valid_decompose_CONV">
<pre>@@CODEREF "valid_decompose_CONV"
</pre>
<***/DESC>

<***DESC NAME="pred_eq_decompose_CONV">
<pre>@@CODEREF "pred_eq_decompose_CONV"
</pre>
<***/DESC>

<***DESC NAME="tactics">
<pre>@@CODEREF "tactics"
</pre>
<***/DESC>

--------------------------------------------------------------- *)

structure predLib = struct


local 

open HolKernel basicHol90Lib predTheory
infix THEN THENL THENC ORELSE ORELSEC THEN_TCL ORELSE_TCL ## |-> 
infixr --> 

in

val pred_defs = [FF_def, TT_def, pNOT_def, pAND_def, pOR_def, 
                 pIMP_def, pEQUAL_def, RES_qAND_def, RES_qOR_def, valid_def ] ;

(* --- <***CODE NAME="valid_decompose_CONV"> --- *)

val valid_decompose_CONV =
	PURE_REWRITE_CONV pred_defs
	THENC
	REPEATC (CHANGED_CONV (DEPTH_CONV BETA_CONV))				;

(* --- <***/CODE> --- *)

(* --- <***CODE NAME="pred_eq_decompose_CONV"> --- *)

val pred_eq_decompose_CONV =
	FUN_EQ_CONV
	THENC
	valid_decompose_CONV							;

(* --- <***/CODE> --- *)

(* --- <***CODE NAME="tactics"> --- *)

val valid_decompose_TAC = CONV_TAC valid_decompose_CONV				;
val pred_eq_decompose_TAC = CONV_TAC pred_eq_decompose_CONV			;
val valid_PROVE_TAC = valid_decompose_TAC THEN bossLib.PROVE_TAC []		;
val pred_eq_PROVE_TAC = pred_eq_decompose_TAC THEN bossLib.PROVE_TAC []		;

(* --- <***/CODE> --- *)

end
end
