(*--------------------------------------------------------------------------


<***MODULE>predTheory<***/MODULE>

<***DESC NAME="ROOT">
Although HOL formulas are predicate logic formulas (predicates), we
can within HOL represents the semantics of predicates. Semantically, a
predicate is a function from the domain of the predicate to
bool. Having a representation of predicates as HOL values is useful as
it enables us to use HOL to analyze their properties. Note that we
cannot use HOL to, for example, make a syntactical analysis on HOL's
own formulas because the syntactical structure of a HOL formula is not
a HOL value, and hence cannot be subjected to HOL's own logic. <p>

This theory provides a set of standard predicate operators in which
predicates are represented semantically as functions from some domain
to bool.
<***/DESC>

<***AUTHOR> 
@@NAME Wishnu Prasetya
@@EMAIL wishnu@cs.uu.nl
@@EMAIL conjurerPrime@yahoo.com
@@URL www.cs.uu.nl/~wishnu
<***/AUTHOR>

<***VERSION>
@@DATE Feb 2002 on Hol2.00 Taupo 6
@@DATE Sept 2001
<***/VERSION>


--------------------------------------------------------------------------*)

open Parse HolKernel basicHol90Lib 

infix THEN THENL THENC ORELSE ORELSEC THEN_TCL ORELSE_TCL ## |-> 

infixr --> 
  
val _ = new_theory "pred" 


(* load "bossLib" ; *)

(*---------------------------------------------------------------------------

<***DESC NAME="Supported Operators">

<blockquote>
<pre>TT              =  (\s. T)
FF              =  (\s. F)
NOT p           =  (\s. ~p s)
p AND q		=  (\s. p s /\ q s)
p OR q          =  (\s. p s \/ q s)
p IMP q         =  (\s. p s ==> q s)
p EQUAL q       =  (\s. p s = q s)
(!!i::P. Q i)   =  (\s. (!i. P i ==> Q i s))
(??i::P. Q i)   =  (\s. (?i. P i /\  Q i s))
valid p         = (!s. p s)
</pre>
</blockquote>

The priorites and associativities of the infixes are as follows:

<blockquote>
<PRE>AND	infixl 720
OR	infixl 710
IMP	infixr 705
EQUAL	infixr 701
</PRE>
</blockquote>
<***/DESC>

---------------------------------------------------------------------------*)

val TT_def 
	= 
	new_definition ("TT_def", 
	--`TT = \s. T`--					)	;

val FF_def 
	= 
	new_definition ("FF_def",
	--`FF = \s.F`--						)	;

val pNOT_def 
	= 
	new_definition ("pNOT_def", 
	--`NOT p 	= \s. ~p s`--				)	;

val pAND_def 
	= 
	new_infixl_definition ("pAND_def",
	--`AND p q 	= \s. p s /\ q s`--,
	720							) 	;

val pOR_def  
	= 
	new_infixl_definition ("pOR_def",
	--`OR p q 	= \s. p s \/ q s`--,
 	710							)	;

val pIMP_def 
	= 
	new_infixr_definition ("pIMP_def",
	--`IMP p q 	= \s. p s ==> q s`--,
	705							) 	;

(* I actually want EQUAL to be a nofix! But I can't figure out how to do it. *)

val pEQUAL_def 
	= 
	new_infixr_definition ("pEQUAL_def",
	--`EQUAL p q = \s. p s = q s`--,
	701							) 	;

val RES_qAND_def 
	=
	new_definition ("RES_qAND_def",
	--`RES_qAND W P = (\s. !i. W i ==> P i s)`--		) 	;

val RES_qOR_def  
	= 
	new_definition ("RES_qOR_def",
	--`RES_qOR W P  = (\s. ?i. W i /\  P i s)`--		) 	;

val valid_def = new_definition ("valid_def",
	--`valid p = !s:'a. p s`--				)	;

(* Parsing stuff: Setting up the fixity and biders *)

val _ = Psyntax.new_binder ("!!", Type `:('b->bool)->'a->bool`)  ;
val _ = Psyntax.new_binder ("??", Type `:('b->bool)->'a->bool`)  ;

val _ = associate_restriction ("!!", "RES_qAND") ;
val _ = associate_restriction ("??", "RES_qOR") ;


val _ = export_theory() ; 
