(*====================================================================

<***MODULE>uprogTheory<***/MODULE>

<***DESC NAME="ROOT"> 
This theory defines some notions on the UNITY program level.
<***/DESC>

<***AUTHOR> 
@@NAME Wishnu Prasetya
@@EMAIL wishnu@cs.uu.nl
@@EMAIL conjurerPrime@yahoo.com
@@URL www.cs.uu.nl/~wishnu
<***/AUTHOR>

<***VERSION>
@@DATE Feb 2002 on Hol2.00 Taupo 6
@@DATE Sept 2001
<***/VERSION>

====================================================================*)

open HolKernel Parse basicHol90Lib ;
infix THEN THENL THENC ORELSE ORELSEC THEN_TCL ORELSE_TCL ## |-> ;
infixr --> ;

open predTheory actionTheory ;

val _ = new_theory "uprog" ;

(*====================================================================
<***DESC name="UNITY program">

A UNITY program is a set of actions. The actions are assumed to be
terminating and atomics. The set will be represented by a function of
type <tt>'a->bool</tt>.

<***/DESC>

<***DESC name="Parallel Composition">

Parallel composition between two programs:
<blockquote>
<PRE>PAR_def : |- !P Q.  P || Q  =  P OR Q</PRE>
</blockquote>
Treated as Infixl 510.

<p>We also have a quantification form, e.g. as in:

<blockquote>
<PRE>(PAR i::V. Program i)</PRE>
</blockquote>

The binder <tt>PAR</tt> is associated with <tt>RES_PAR</tt>, which is
equal to <tt>RES_qOR</tt>. 

<***/DESC>

====================================================================*)

val PAR_def
	=
	new_infixl_definition ("PAR_def",
	--`$|| P Q = P OR Q`--,
	510						)	;

val RES_PAR_def
	=
	new_definition ("RES_PAR_def", 
	--`RES_PAR = RES_qOR`--) ;

val _ = Psyntax.new_binder ("PAR", Type `:('b->bool)->'a->bool`)  ;

val _ = associate_restriction ("PAR", "RES_PAR") ;

(*====================================================================

<***DESC name="STABLE">

<blockquote>
<pre>STABLE_def : |- !P J. STABLE P J = !a. P a ==> HOA J a J</pre>
</blockquote>

<***/DESC>
====================================================================*)

val STABLE_def
	=
	new_definition ("STABLE_def",
	--`STABLE P J = !a. P a ==> HOA J a J`--	)	;	

(*
val _ = add_rule 
	{ term_name 	= "STABLE",
	  fixity 	= TruePrefix 500,
	  pp_elements	= [TOK "PROP", BreakSpace(1,2), TM, BreakSpace(1,0),
			   TOK "|--", BreakSpace(1,0),
			   BeginFinalBlock(PP.CONSISTENT,2),
			   TOK "STABLE", BreakSpace(1,0)],
	  paren_style	= OnlyIfNecessary,
	  block_style	= (AroundEachPhrase, (PP.INCONSISTENT, 0))
	} ;
*)

val _ = export_theory() ;
